#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2020/06/02 11:09

from collections import defaultdict, Counter
import copy
import re
import sys

import pandas as pd
from more_itertools import unique_everseen

from pyxllib.prog.newbie import typename
from pyxllib.algo.pupil import natural_sort_key
from pyxllib.text.pupil import shorten, east_asian_shorten


def dataframe_str(df, *args, ambiguous_as_wide=None, shorten=True):
    """输出DataFrame
    DataFrame可以直接输出的，这里是增加了对中文字符的对齐效果支持

    :param df: DataFrame数据结构
    :param args: option_context格式控制
    :param ambiguous_as_wide: 是否对①②③这种域宽有歧义的设为宽字符
        win32平台上和linux上①域宽不同，默认win32是域宽2，linux是域宽1
    :param shorten: 是否对每个元素提前进行字符串化并控制长度在display.max_colwidth以内
        因为pandas的字符串截取遇到中文是有问题的，可以用我自定义的函数先做截取
        默认开启，不过这步比较消耗时间

    >> df = pd.DataFrame({'哈哈': ['a'*100, '哈\n①'*10, 'a哈'*100]})
                                                        哈哈
        0  aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa...
        1   哈 ①哈 ①哈 ①哈 ①哈 ①哈 ①哈 ①哈 ①哈 ①...
        2  a哈a哈a哈a哈a哈a哈a哈a哈a哈a哈a哈a哈a哈a哈a哈a...
    """
    import pandas as pd

    if ambiguous_as_wide is None:
        ambiguous_as_wide = sys.platform == 'win32'
    with pd.option_context('display.unicode.east_asian_width', True,  # 中文输出必备选项，用来控制正确的域宽
                           'display.unicode.ambiguous_as_wide', ambiguous_as_wide,
                           'max_columns', 20,  # 最大列数设置到20列
                           'display.width', 200,  # 最大宽度设置到200
                           *args):
        if shorten:  # applymap可以对所有的元素进行映射处理，并返回一个新的df
            df = df.applymap(lambda x: east_asian_shorten(str(x), pd.options.display.max_colwidth))
        s = str(df)
    return s


class TypeConvert:
    @classmethod
    def dict2list(cls, d: dict, *, nsort=False):
        """ 字典转n*2的list

        :param d: 字典
        :param nsort:
            True: 对key使用自然排序
            False: 使用d默认的遍历顺序
        :return:
        """
        ls = list(d.items())
        if nsort:
            ls = sorted(ls, key=lambda x: natural_sort_key(str(x[0])))
        return ls

    @classmethod
    def dict2df(cls, d):
        """dict类型转DataFrame类型"""
        name = typename(d)
        if isinstance(d, Counter):
            li = d.most_common()
        else:
            li = cls.dict2list(d, nsort=True)
        return pd.DataFrame.from_records(li, columns=(f'{name}-key', f'{name}-value'))

    @classmethod
    def list2df(cls, li):
        if li and isinstance(li[0], (list, tuple)):  # 有两维时按表格显示
            df = pd.DataFrame.from_records(li)
        else:  # 只有一维时按一列显示
            df = pd.DataFrame(pd.Series(li), columns=(typename(li),))
        return df

    @classmethod
    def try2df(cls, arg):
        """尝试将各种不同的类型转成dataframe"""
        if isinstance(arg, dict):
            df = cls.dict2df(arg)
        elif isinstance(arg, (list, tuple)):
            df = cls.list2df(arg)
        elif isinstance(arg, pd.Series):
            df = pd.DataFrame(arg)
        else:
            df = arg
        return df


class NestedDict:
    """ 字典嵌套结构相关功能

    TODO 感觉跟 pprint 的嵌套识别美化输出相关，可能有些代码是可以结合简化的~~
    """

    @classmethod
    def has_subdict(cls, data, include_self=True):
        """是否含有dict子结构
        :param include_self: 是否包含自身，即data本身是一个dict的话，也认为has_subdict是True
        """
        if include_self and isinstance(data, dict):
            return True
        elif isinstance(data, (list, tuple, set)):
            for v in data:
                if cls.has_subdict(v):
                    return True
        return False

    @classmethod
    def to_html_table(cls, data, max_items=10):
        """ 以html表格套表格的形式，展示一个嵌套结构数据

        :param data: 数据
        :param max_items: 项目显示上限，有些数据项目太多了，要精简下
            设为假值则不设上限
        :return:

        TODO 这个速度有点慢，怎么加速？
        """

        def tohtml(d):
            if max_items:
                df = TypeConvert.try2df(d)
                if len(df) > max_items:
                    n = len(df)
                    return df[:max_items].to_html(escape=False) + f'... {n - 1}'
                else:
                    return df.to_html(escape=False)
            else:
                return TypeConvert.try2df(d).to_html(escape=False)

        if not cls.has_subdict(data):
            res = str(data)
        elif isinstance(data, dict):
            if isinstance(data, Counter):
                d = data
            else:
                d = dict()
                for k, v in data.items():
                    if cls.has_subdict(v):
                        v = cls.to_html_table(v, max_items=max_items)
                    d[k] = v
            res = tohtml(d)
        else:
            li = [cls.to_html_table(x, max_items=max_items) for x in data]
            res = tohtml(li)

        return res.replace('\n', ' ')


class KeyValuesCounter:
    """ 各种键值对出现次数的统计
    会递归找子字典结构，但不存储结构信息，只记录纯粹的键值对信息

    应用场景：对未知的json结构，批量读取后，显示所有键值对的出现情况
    """

    def __init__(self):
        self.kvs = defaultdict(Counter)

    def add(self, data, max_value_length=100):
        """
        :param max_value_length: 添加的值，进行截断，防止有些值太长
        """
        if not NestedDict.has_subdict(data):
            return
        elif isinstance(data, dict):
            for k, v in data.items():
                if NestedDict.has_subdict(v):
                    self.add(v)
                else:
                    self.kvs[k][shorten(str(v), max_value_length)] += 1
        else:  # 否则 data 应该是个可迭代对象，才可能含有dict
            for x in data:
                self.add(x)

    def to_html_table(self, max_items=10):
        return NestedDict.to_html_table(self.kvs, max_items=max_items)


class JsonStructParser:
    """ 类json数据格式的结构解析

    【名称定义】
    item: 一条类json的数据条目
    path: 用类路径的格式，表达item中某个数值的索引。例如
        /a/b/3/c: 相当于 item['a']['b'][3]['c']
        有一些特殊的path，例如容器类会以/结尾： /a/
        以及一般会带上数值的类型标记，区分度更精确：/a/=dict
    pathx: 泛指下述中某种格式
        pathlist: list, 一条item对应的扁平化的路径
        pathstr/struct: paths拼接成一个str
        pathdict: paths重新组装成一个dict字典(未实装，太难写，性价比也低)
    """

    default_cfg = {'include_container': True,  # 包含容器（dict、list）的路径
                   'value_type': True,  # 是否带上后缀：数值的类型
                   # 可以输入一个自定义的路径合并函数 path,type=merge_path(path,type)。
                   # 一般是字典中出现不断变化的数值id，格式不统一，使用一定的规则，可以将path几种相近的冗余形式合并。
                   # 也可以设True，默认会将数值类统一为0。
                   'merge_path': False,
                   }

    @classmethod
    def _get_item_path_types(cls, item, prefix=''):
        """
        :param item: 类json结构的数据，可以含有类型： dict, list(tuple), int, str, bool, None
            结点类型
                其中 dict、list称为 container 容器类型
                其他int、str称为数值类型
            结构
                item 可以看成一个树形结构
                其中数值类型可以视为叶子结点，其他容器类是非叶子结点

        瑕疵
        1、如果key本身带有"/"，会导致混乱
        2、list的下标转为0123，和字符串类型的key会混淆，和普通的字典key也会混淆
        """
        path_types = []
        if isinstance(item, dict):
            path_types.append([prefix + '/', 'dict'])
            for k in sorted(item.keys()):  # 实验表明，在这里对字典的键进行排序就行，最后总的paths不要排序，不然结构会乱
                v = item[k]
                path_types.extend(cls._get_item_path_types(v, f'{prefix}/{k}'))
        elif isinstance(item, (list, tuple)):
            path_types.append([prefix + '/', type(item).__name__])
            for k, v in enumerate(item):
                path_types.extend(cls._get_item_path_types(v, f'{prefix}/{k}'))
        else:
            path_types.append([prefix, type(item).__name__])
        return path_types

    @classmethod
    def get_item_pathlist(cls, item, prefix='', **kwargs):
        """ 获得字典的结构标识
        """
        # 1 底层数据
        cfg = copy.copy(cls.default_cfg)
        cfg.update(kwargs)
        paths = cls._get_item_path_types(item, prefix)

        # 2 配置参数
        if cfg['merge_path']:
            if callable(cfg['merge_path']):
                func = cfg['merge_path']
            else:
                def func(p, t):
                    return re.sub(r'\d+', '0', p), t

            # 保序去重
            paths = list(unique_everseen(map(lambda x: func(x[0], x[1]), paths)))

        if not cfg['include_container']:
            paths = [pt for pt in paths if (pt[0][-1] != '/')]

        if cfg['value_type']:
            paths = ['='.join(pt) for pt in paths]
        else:
            paths = [pt[0] for pt in paths]

        return paths

    @classmethod
    def get_item_pathstr(cls, item, prefix='', **kwargs):
        paths = cls.get_item_pathlist(item, prefix, **kwargs)
        return '\n'.join(paths)

    @classmethod
    def get_items_struct2cnt(cls, items, **kwargs):
        # 1 统计每种结构出现的次数
        struct2cnt = Counter()
        for item in items:
            pathstr = cls.get_item_pathstr(item, **kwargs)
            struct2cnt[pathstr] += 1
        # 2 按照从多到少排序
        struct2cnt = Counter(dict(sorted(struct2cnt.items(), key=lambda item: -item[1])))
        return struct2cnt

    @classmethod
    def get_items_structdf(cls, items, **kwargs):
        """ 分析一组题目里，出现了多少种不同的json结构 """
        # 1 获取原始数据，初始化
        struct2cnt = cls.get_items_struct2cnt(items, **kwargs)
        m = len(struct2cnt)

        # 2 path2cnt
        path2cnt = Counter()
        for struct in struct2cnt.keys():
            path2cnt.update({path: struct2cnt[struct] for path in struct.splitlines()})
        paths = sorted(path2cnt.keys(), key=lambda path: re.split(r'/=', path))
        path2cnt = {path: path2cnt[path] for path in paths}

        # 3 生成统计表
        ls = []
        columns = ['path', 'total'] + [f'struct{i}' for i in range(1, m + 1)]
        for path, cnt in path2cnt.items():
            row = [path, cnt]
            for struct, cnt in struct2cnt.items():
                t = cnt if path in struct else 0
                row.append(t)
            ls.append(row)

        df = pd.DataFrame.from_records(ls, columns=columns)
        return df

    @classmethod
    def get_itemgroups_structdf(cls, itemgroups, **kwargs):
        """ 分析不同套数据间的json结构区别

        这里为了减少冗余开发，直接复用get_items_structdf
            虽然会造成一些冗余功能，
        """
        # 1 统计所有paths出现情况
        n = len(itemgroups)
        d = dict()
        for i, gs in enumerate(itemgroups):
            for x in gs:
                paths = cls.get_item_pathlist(x, **kwargs)
                for p in paths:
                    if p not in d:
                        d[p] = [0] * n
                    d[p][i] += 1

        # 排序
        paths = sorted(d.keys(), key=lambda path: re.split(r'/=', path))

        # 2 统计表
        ls = []
        columns = ['path', 'total'] + [f'group{i}' for i in range(1, n + 1)]
        for path in paths:
            vals = d[path]
            row = [path, sum(vals)] + vals
            ls.append(row)

        df = pd.DataFrame.from_records(ls, columns=columns)
        return df



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2022/02/15 09:51

import os
import shutil
from tarfile import TarFile
# 这个包里的ZipFile是拷贝后修改过的，不影响标准库zipfile里的功能
import pyxllib.file.packlib.zipfile as zipfile
from pyxllib.file.packlib.zipfile import ZipFile

from pyxllib.file.specialist import XlPath, reduce_dir_depth
from pyxllib.prog.pupil import inject_members

""" 问题：py官方的ZipFile解压带中文的文件会乱码

解决办法及操作流程
1、定期将官方的zipfile.py文件更新到pyxllib.file.packlib.zipfile
2、将代码中两处cp437都替换为gbk
"""


def unpack_zipfile(filename, extract_dir):
    """ 为了修复zipfile底层的中文解压乱码问题，修改了shutil._UNPACK_FORMATS的底层功能
    """
    from shutil import ReadError

    zip = zipfile.ZipFile(filename)
    if not zipfile.is_zipfile(filename):
        raise ReadError("%s is not a zip file" % filename)
    try:
        for info in zip.infolist():
            name = info.filename

            # don't extract absolute paths or ones with .. in them
            if name.startswith('/') or '..' in name:
                continue

            target = os.path.join(extract_dir, *name.split('/'))
            if not target:
                continue

            os.makedirs(XlPath(target).parent, exist_ok=True)
            if not name.endswith('/'):
                # file
                data = zip.read(info.filename)
                f = open(target, 'wb')
                try:
                    f.write(data)
                finally:
                    f.close()
                    del data
    finally:
        zip.close()


# 解决unzip中文乱码问题
shutil._UNPACK_FORMATS['zip'] = (['.zip'], unpack_zipfile, [], "ZIP file")


def _parse_depth(names):
    """ 判断有几层目录都只有独立的一个目录

    :param names: 一组相对路径名
    :return: 如果直接目录下文件已经不唯一，返回0

    >>> _parse_depth(['a'])
    1
    >>> _parse_depth(['a', 'a/1'])
    2
    >>> _parse_depth(['a', 'a/b', 'a/b/1', 'a/b/2'])
    2
    >>> _parse_depth(['a', 'a/b', 'a/b/1', 'a/c/3'])
    1
    """
    parents = []
    for name in names:
        p = XlPath(name)
        ps = [x.as_posix() for x in p.parents]
        parents.append(ps[-2::-1] + [p.as_posix()])
    i = 0
    while len({x[i] for x in parents if len(x) > i}) == 1:
        i += 1
    return i


def _unpack_base(packfile, namelist, format=None, extract_dir=None, wrap=0):
    """ 解压

    :param packfile: 压缩包本身的文件路径名称，在wrap=1时可能会用到，用来添加一个同名目录
    :param namelist: 压缩包里所有文件的相对路径
    :param format: 压缩包类型，比如zip、tar，可以不输入，此时直接采用shutil.unpack_archive里的判断机制
    :param extract_dir: 解压目标目录，可以不输入，会自动判断
        自动解压规则：压缩包里目录不唯一，则以压缩包本身的名称创建一个目录存放
            否则以压缩包里的目录名为准
    :param wrap: 解压时倾向的处理机制
        0，默认，不进行任何处理
        1，如果压缩包里一级目录下不止一个对象，使用压缩包本身的名称再包一层
        -1，如果压缩包里一级目录下只有一个目录，则丢弃这个目录，只取里面的内容出来
            同理，可能这一个目录里的文件，还是只有一个目录，此时支持设置-2、-3等解开上限
            也就-1比较常用，-2、-3等很少见了
        TODO 还有一种解压机制，只取里面的部分文件，有空可以继续调研扩展下
    :return: 无论怎样，最后都是放到一个目录里，返回改目录路径
    """

    # 1 默认解压目录是self所在目录，注意这点跟shutil.unpack_archive的当前工作目录不同
    p0 = XlPath(packfile)
    if extract_dir is None:
        extract_dir = p0.parent

    depth = _parse_depth(namelist)

    # 2 是否要加目录层级，或者去掉目录层级
    if wrap == 1 and depth == 0:
        # 压缩包里直接目录下有多个文件，且指定了wrap，则加一层目录
        shutil.unpack_archive(packfile, extract_dir / p0.stem, format)
    elif wrap < 0 < depth:
        # 压缩包有depth层冗余目录，并且指定wrap要解开这几层目录
        # 先正常解压（每种压缩包格式内部处理机制都有点不太一样，所以先解压，然后用目录文件功能处理更合适）
        shutil.unpack_archive(packfile, extract_dir, format)
        reduce_dir_depth(extract_dir, unwrap=-wrap)
    else:
        # 正常解压
        shutil.unpack_archive(packfile, extract_dir, format)


class XlZipFile(ZipFile):

    def infolist2(self, prefix=None, zipinfo=True):
        """>> self.infolist2()  # getinfo的多文件版本
             1           <ZipInfo filename='[Content_Types].xml' compress_type=deflate file_size=1495 compress_size=383>
             2                    <ZipInfo filename='_rels/.rels' compress_type=deflate file_size=590 compress_size=243>
            ......
            20            <ZipInfo filename='word/fontTable.xml' compress_type=deflate file_size=1590 compress_size=521>
            21               <ZipInfo filename='docProps/app.xml' compress_type=deflate file_size=720 compress_size=384>

        :param prefix:
            可以筛选文件的前缀，例如“word/”可以筛选出word目录下的
        :param zipinfo:
            返回的list每个元素是zipinfo数据类型
        """
        ls = self.infolist()
        if prefix:
            ls = list(filter(lambda t: t.filename.startswith(prefix), ls))
        if not zipinfo:
            ls = list(map(lambda x: x.filename, ls))
        return ls

    def unpack(self, extract_dir=None, format='zip', wrap=0):
        _unpack_base(self.filename, self.namelist(), format, extract_dir, wrap)


class XlTarFile(TarFile):

    def unpack(self, extract_dir=None, format='tar', wrap=0):
        _unpack_base(self.name, self.getnames(), format, extract_dir, wrap)


inject_members(XlZipFile, ZipFile)
inject_members(XlTarFile, TarFile)


def unpack_archive(filename, extract_dir=None, format=None, *, wrap=0):
    """ 对shutil.unpack_archive的扩展，增加了一个wrap的接口功能 """
    if format is None:
        format = shutil._find_unpack_format(str(filename).lower())

    if format.endswith('zip'):
        XlZipFile(filename).unpack(extract_dir, format, wrap)
    elif format.endswith('tar'):
        XlTarFile(filename).unpack(extract_dir, format, wrap)
    else:
        # 其他还没扩展的格式，不支持wrap功能，但仍然可以使用shutil标准的接口解压
        shutil.unpack_archive(filename, extract_dir, format)


def compress_to_zip(source_path, target_zip_path=None, wrap=None):
    """ 压缩指定的文件或文件夹为ZIP格式。

    :param str source_path: 要压缩的文件或文件夹路径
    :param str target_zip_path: 目标ZIP文件路径（可选）
    :param str wrap: 在ZIP文件内部创建的目录名，所有内容将被放在这个目录下
    """
    # 根据输入路径生成默认的目标ZIP文件路径
    if target_zip_path is None:
        # 获取输入路径的基本名称（文件名或文件夹名）
        base_name = os.path.basename(source_path)
        # 为基本名称添加'.zip'后缀
        target_zip_path = os.path.join(os.path.dirname(source_path), f"{base_name}.zip")

    # 创建一个ZipFile对象来写入压缩文件
    with XlZipFile(target_zip_path, 'w', zipfile.ZIP_DEFLATED) as zipf:
        # 如果是文件夹，则遍历文件夹中的所有文件和子文件夹
        if os.path.isdir(source_path):
            for root, _, files in os.walk(source_path):
                for file in files:
                    file_path = os.path.join(root, file)
                    # 计算文件在ZIP中的相对路径
                    arcname = os.path.relpath(file_path, source_path)
                    if wrap:
                        arcname = os.path.join(wrap, arcname)
                    zipf.write(file_path, arcname)
        # 如果是文件，则直接将文件添加到ZIP中
        elif os.path.isfile(source_path):
            arcname = os.path.basename(source_path)
            if wrap:
                arcname = os.path.join(wrap, arcname)
            zipf.write(source_path, arcname)

    return target_zip_path



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2021/06/03 23:04

""" 统计方面的功能

主要是pandas、表格运算
"""

import sys
from collections import defaultdict, Counter

import pandas as pd

from pyxllib.prog.pupil import dprint, typename
from pyxllib.file.specialist import XlPath


def treetable(childreds, parents, arg3=None, nodename_colname=None):
    """ 输入childres子结点id列表，和parents父结点id列表

    两个列表长度必须相等
    文档：http://note.youdao.com/noteshare?id=126200f45d301fcb4364d06a0cae8376

    有两种调用形式
    >> treetable(childreds, parents)  --> DataFrame  （新建df）
    >> treetable(df, child_colname, parent_colname)  --> DataFrame （修改后的df）

    返回一个二维列表
        新的childreds （末尾可能回加虚结点）
        新的parents
        函数会计算每一行childred对应的树排序后的排序编号order
        以及每个节点深度depth

    >> ls1 = [6, 2, 4, 5, 3], ls2 = [7, 1, 2, 2, 1], treetable(ls1, ls2)
          child_id   parent_id   depth     tree_order    tree_struct
        5        7     root        1           1         = = 7
        0        6        7        2           2         = = = = 6
        6        1     root        1           3         = = 1
        1        2        1        2           4         = = = = 2
        2        4        2        3           5         = = = = = = 4
        3        5        2        3           6         = = = = = = 5
        4        3        1        2           7         = = = = 3
    """
    # 0 参数预处理
    if isinstance(childreds, pd.DataFrame):
        df = childreds
        child_colname = parents
        parent_colname = arg3
        if not arg3: raise TypeError
        childreds = df[child_colname].tolist()
        parents = df[parent_colname].tolist()
    else:
        df = None

    # 1 建立root根节点，确保除了root其他结点都存在记录
    lefts = set(parents) - set(childreds)  # parents列中没有在childreds出现的结点
    cs, ps = list(childreds), list(parents)

    if len(lefts) == 0:
        # b_left为空一定有环，b_left不为空也不一定是正常的树
        raise ValueError('有环，不是树结构')
    elif len(lefts) == 1:  # 只有一个未出现的结点，那么它既是根节点
        root = list(lefts)[0]
    else:  # 多个父结点没有记录，则对这些父结点统一加一个root父结点
        root = 'root'
        allnode = set(parents) | set(childreds)  # 所有结点集合
        while root in allnode: root += '-'  # 一直在末尾加'-'，直到这个结点是输入里未出现的
        # 添加结点
        lefts = list(lefts)
        lefts.sort(key=lambda x: parents.index(x))
        for t in lefts:
            cs.append(t)
            ps.append(root)

    n = len(cs)
    depth, tree_order, len_childs = [-1] * n, [-1] * n, [0] * n

    # 2 构造父结点-孩子结点的字典dd
    dd = defaultdict(list)
    for i in range(n): dd[ps[i]] += [i]

    # 3 dfs
    cnt = 1

    def dfs(node, d):
        """找node的所有子结点"""
        nonlocal cnt
        for i in dd.get(node, []):
            tree_order[i], depth[i], len_childs[i] = cnt, d, len(dd[cs[i]])
            cnt += 1
            dfs(cs[i], d + 1)

    dfs(root, 1)

    # 4 输出格式
    tree_struct = list(map(lambda i: f"{'_ _ ' * depth[i]}{cs[i]}" + (f'[{len_childs[i]}]' if len_childs[i] else ''),
                           range(n)))

    if df is None:
        ls = list(zip(cs, ps, depth, tree_order, len_childs, tree_struct))
        df = pd.DataFrame.from_records(ls, columns=('child_id', 'parent_id',
                                                    'depth', 'tree_order', 'len_childs', 'tree_struct'))
    else:
        k = len(df)
        df = df.append(pd.DataFrame({child_colname: cs[k:], parent_colname: ps[k:]}), sort=False, ignore_index=True)
        if nodename_colname:
            tree_struct = list(
                map(lambda i: f"{'_ _ ' * depth[i]}{cs[i]} {df.iloc[i][nodename_colname]}"
                              + (f'[{len_childs[i]}]' if len_childs[i] else ''), range(n)))
        df['depth'], df['tree_order'], df['len_childs'], df['tree_struct'] = depth, tree_order, len_childs, tree_struct
    df.sort_values('tree_order', inplace=True)  # 注意有时候可能不能排序，要维持输入时候的顺序
    return df


def treetable_flatten(df, *, reverse=False, childid_colname='id', parentid_colname='parent_id', format_colname=None):
    """ 获得知识树横向展开表：列为depth-3, depth-2, depth-1，表示倒数第3级、倒数第2级、倒数第1级

    :param df: DataFrame数据
    :param reverse:
        False，正常地罗列depth1、depth2、depth3...等结点信息
        True，反向列举所属层级，即显示倒数第1层parent1，然后是倒数第2层parent2...
    :param childid_colname: 孩子结点列
    :param parentid_colname: 父结点列
    :param format_colname: 显示的数值
        None，默认采用 childid_colname 的值
        str，某一列的名称，采用那一列的值（可以实现设置好格式）
    :return:
    """
    # 1 构造辅助数组
    if format_colname is None: format_colname = parentid_colname
    parentid = dict()  # parentid[k] = v， 存储结点k对应的父结点v
    nodeval = dict()  # nodeval[k] = v，  存储结点k需要显示的数值情况
    if len(df[df.index.duplicated()]):
        dprint(len(set(df.index)), len(df.index))  # 有重复index
        raise ValueError

    for idx, row in df.iterrows():
        parentid[row[childid_colname]] = row[parentid_colname]
        nodeval[row[childid_colname]] = str(row[format_colname])

    # 2 每个结点往上遍历出所有父结点
    parents = []
    for idx, row in df.iterrows():
        ps = [nodeval[row[childid_colname]]]  # 包含结点自身的所有父结点名称
        p = row[parentid_colname]
        while p in parentid:
            ps.append(nodeval[p])
            p = parentid[p]
        parents.append(ps)
    num_depth = max(map(len, parents), default=0)

    # 3 这里可以灵活调整最终要显示的格式效果
    df['parents'] = parents
    if reverse:
        for j in range(num_depth, 0, -1): df[f'depth-{j}'] = ''
        for idx, row in df.iterrows():
            for j in range(1, len(row.parents) + 1):
                df.loc[idx, f'depth-{j}'] = row.parents[j - 1]
    else:
        for j in range(num_depth): df[f'depth{j}'] = ''
        for idx, row in df.iterrows():
            for j in range(len(row.parents)):
                df.loc[idx, f'depth{j}'] = row.parents[-j - 1]
    df.drop('parents', axis=1, inplace=True)
    return df


def write_dataframes_to_excel(outfile, dataframes, order_mode='序号'):
    """ 将多个DataFrame表格写入一个Excel文件，并添加序号列

    :param str outfile: 输出的Excel文件路径
    :param dict dataframes: 包含要保存的DataFrame的字典，键为sheet名，值为DataFrame
    :param str order_mode: 序号模式，可选值为 'default' 或 '序号'，默认为 '序号'

    >> write_dataframes_to_excel('test.xlsx', {'images': df1, 'annotations': df2})

    # TODO 存成表格后，可以使用openpyxl等库再打开表格精修

    实现上，尽可能在一些常见结构上，进行一些格式美化。但对费常规结构，就保留df默认排版效果，不做特殊处理。
    """
    with pd.ExcelWriter(str(outfile)) as writer:
        head_format = writer.book.add_format({'font_size': 12, 'font_color': 'blue',
                                              'align': 'left', 'valign': 'vcenter'})
        for sheet_name, df in dataframes.items():
            if df.index.nlevels == 1 and df.columns.nlevels == 1:
                if order_mode == '序号':
                    # 写入带有序号列的数据表格
                    if '序号' not in df.columns:
                        df = df.copy()
                        df.insert(0, '序号', range(1, len(df) + 1))
                else:
                    df = df.reset_index()
                    df.columns = ['_index'] + list(df.columns[1:])
                df.to_excel(writer, sheet_name=sheet_name, freeze_panes=(1, 0), index=False)
            else:
                # 写入普通的数据表格
                df.to_excel(writer, sheet_name=sheet_name, freeze_panes=(1, df.index.nlevels))

            # 设置表头格式
            if df.columns.nlevels == 1:
                start = df.index.nlevels
                if start == 1:
                    start = 0
                for col_num, value in enumerate(df.columns, start=start):
                    writer.sheets[sheet_name].write(0, col_num, value, head_format)


def read_dataframes_from_excel(infile):
    """ 从Excel文件读取多个DataFrame表格

    :param str infile: Excel文件路径
    :return: 包含读取的DataFrame的字典，键为工作表名，值为DataFrame
    :rtype: dict

    注意这个函数不太适用于与读取多级index和多级columns的情况，建议遇到这种情况，手动读取，
        read_excel可以设置header=[0,1]、index=[0,1,2]的形式来定制表头所在位置。
    """
    dataframes = {}
    with pd.ExcelFile(infile) as xls:
        sheet_names = xls.sheet_names
        for sheet_name in sheet_names:
            df = pd.read_excel(xls, sheet_name=sheet_name)
            if '_index' in df.columns:
                df = df.drop('_index', axis=1)
            dataframes[sheet_name] = df
    return dataframes


def update_dataframes_to_excel(outfile, dataframes, order_mode='序号'):
    """ 更新xlsx文件中的sheets数据 """
    outfile = XlPath(outfile)
    if outfile.is_file():
        data = read_dataframes_from_excel(outfile)
    else:
        data = {}
    data.update(dataframes)
    write_dataframes_to_excel(outfile, data, order_mode)


def xlpivot(df, index=None, columns=None, values=None):
    """ 对pandas进行封装的数据透视表功能

    :param df: 数据表
    :param index: 行划分方式
    :param columns: 列划分方式
    :param values: 显示的值
        Callable[items, value]：输出一个函数
    :return: 数据透视表的表格

    使用示例：
    def func(items):
        x = items.iloc[0]
        return f'{x["precision"]:.0f}，{x["recall"]:.0f}，{x["hmean"]:.2f}，{x["fps"]}'
    >> df2 = xlpivot(df, ['model_type'], ['dataset', 'total_frame'], {'precision，recall，hmean，fps': func})
    """

    # 1 将分组的格式标准化
    def reset_groups(keys):
        if isinstance(keys, (list, tuple)):
            return list(keys)
        elif keys:
            return [keys]
        else:
            return []

    index_, columns_ = reset_groups(index), reset_groups(columns)

    # 2 目标值的格式标准化
    if callable(values):
        values_ = {'values': values}
    elif isinstance(values, dict):
        values_ = values
    else:
        raise TypeError

    # 3 分组
    keys = index_ + columns_
    dfgp = df.groupby(keys)
    data = defaultdict(list)
    for ks, items in dfgp:
        # 要存储分组（keys）相关的值
        if len(keys) == 1:
            data[keys[0]].append(ks)
        else:
            for i, k in enumerate(keys):
                data[k].append(ks[i])
        # 再存储生成的值
        for k, func in values_.items():
            data[k].append(func(items))
    df2 = pd.DataFrame.from_dict(data)

    # 4 可视化表格
    if index and columns:
        view_table = df2.pivot(index=index, columns=columns, values=list(values_.keys()))
    elif index:
        view_table = df2.set_index(index_)
    else:  # 只有columns，没有index
        view_table = df2.set_index(index_).T
    return view_table


def count_key_combinations(df, col_names, count_col_name='count'):
    """ 统计列出的几个列名，各种组合出现的次数

    :param df:
    :param col_names: ['a', 'b', 'c']
    :param count_col_name: 新增的统计出现次数的列名，默认count
    :return: 新的次数统计的df表格

    这个功能跟 SqlCodeGenerator 的 keys_count、one2many很像，是可以代替这两个功能的
    """
    from collections import Counter

    # 0 参数处理
    if isinstance(col_names, str):
        col_names = [col_names]

    # 1 统计每种组合出现的次数
    cols = [df[name] for name in col_names]
    ct = Counter(tuple(zip(*cols)))

    # 2 生成新的df的统计表
    ls = []
    for k, v in ct.most_common():
        ls.append([*k, v])
    df2 = pd.DataFrame.from_records(ls, columns=list(col_names) + [count_col_name])
    return df2


def pareto_accumulate(weights, accuracy=0.01, *, print_mode=False):
    """ 帕累托累计

    可以用来分析主要出现的权重、频次
    二八法则，往往20%的数据，就能解决80%的问题

    :param weights: 一组权重数据
    :param accuracy: 累计精度，当统计到末尾时，可能有大量权重过小的数值
        此时不频繁进行累计权重计算，而是当更新权重累计达到accuracy，才会更新一个记录点
        注意这是全量数据综合的百分比，所以最小更新量就是1%
    :param print_mode: 是否直接展示可视化结果
    :return: [(累计数值数量, ≥当前阈值, 累计权重), ...]

    >>> pareto_accumulate([1, 2, 3, 9, 8, 7, 4, 6, 5])
    [(1, 9, 9), (2, 8, 17), (3, 7, 24), (4, 6, 30), (5, 5, 35), (6, 4, 39), (7, 3, 42), (8, 2, 44), (9, 1, 45)]
    >>> pareto_accumulate([1, 2, 3, 9, 8, 7, 4, 6, 5], 0.1)
    [(1, 9, 9), (2, 8, 17), (3, 7, 24), (4, 6, 30), (5, 5, 35), (7, 3, 42), (9, 1, 45)]
    """
    # 1 基础数据计算
    points = []
    weights = sorted(weights, reverse=True)

    total = sum(weights)
    accuracy = total * accuracy

    acc = 0
    delta = 0
    for i, w in enumerate(weights, start=1):
        acc += w
        delta += w
        if delta >= accuracy:
            points.append((i, w, acc))
            delta = 0
    if delta:
        points.append((len(weights), weights[-1], acc))

    # 2 结果展示
    def fmt(p):
        from pyxllib.text.newbie import format_big_decimal
        ls = [f'{p[0]}it≥{format_big_decimal(p[1])}',
              f'{format_big_decimal(p[2])}({p[2] / total_size:.0%})']
        return '，'.join(map(str, ls))

    total_size = points[-1][2]
    labels = [fmt(p) for p in points]

    pts = [[p[0], p[2]] for p in points]

    if print_mode:
        if sys.platform == 'win32':
            from pyxllib.data.echarts import Line
            from pyxllib.prog.specialist import browser

            x = Line()
            x.add_series('帕累托累积权重', pts, labels=labels, label={'position': 'right'})
            browser(x)
        else:
            print(*labels, sep='\n')

    return pts, labels


class XlDataFrame(pd.DataFrame):
    def check_dtypes(self):
        """ 检查数据类型
        第1列是列名，第2列是原本dtypes显示的类型看，第3列是我扩展的统计的实际数据类型
        """
        d = self.dtypes
        ls = [[k, d[k], Counter([typename(x) for x in v]).most_common()] for k, v in self.iteritems()]
        df = pd.DataFrame.from_records(ls, columns=['name', 'type', 'detail'])
        return df


class ModifiableRow:
    def __init__(self, df, index):
        self.df = df
        self.index = index

    def __getitem__(self, item):
        return self.df.at[self.index, item]

    def __setitem__(self, key, value):
        self.df.at[self.index, key] = value



# Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import sys

__dir__ = os.path.dirname(os.path.abspath(__file__))
sys.path.append(__dir__)
sys.path.append(os.path.abspath(os.path.join(__dir__, '../..')))

os.environ["FLAGS_allocator_strategy"] = 'auto_growth'

import cv2
import copy
import numpy as np
import math
import time
import traceback

import pyxlpr.ppocr.tools.infer.utility as utility
from pyxlpr.ppocr.postprocess import build_post_process
from pyxlpr.ppocr.utils.logging import get_logger
from pyxlpr.ppocr.utils.utility import get_image_file_list, check_and_read_gif

logger = get_logger()


class TextClassifier(object):
    def __init__(self, args):
        self.cls_image_shape = [int(v) for v in args.cls_image_shape.split(",")]
        self.cls_batch_num = args.cls_batch_num
        self.cls_thresh = args.cls_thresh
        postprocess_params = {
            'name': 'ClsPostProcess',
            "label_list": args.label_list,
        }
        self.postprocess_op = build_post_process(postprocess_params)
        self.predictor, self.input_tensor, self.output_tensors, _ = \
            utility.create_predictor(args, 'cls', logger)
        self.use_onnx = args.use_onnx

    def resize_norm_img(self, img):
        imgC, imgH, imgW = self.cls_image_shape
        h = img.shape[0]
        w = img.shape[1]
        ratio = w / float(h)
        if math.ceil(imgH * ratio) > imgW:
            resized_w = imgW
        else:
            resized_w = int(math.ceil(imgH * ratio))
        resized_image = cv2.resize(img, (resized_w, imgH))
        resized_image = resized_image.astype('float32')
        if self.cls_image_shape[0] == 1:
            resized_image = resized_image / 255
            resized_image = resized_image[np.newaxis, :]
        else:
            resized_image = resized_image.transpose((2, 0, 1)) / 255
        resized_image -= 0.5
        resized_image /= 0.5
        padding_im = np.zeros((imgC, imgH, imgW), dtype=np.float32)
        padding_im[:, :, 0:resized_w] = resized_image
        return padding_im

    def __call__(self, img_list):
        img_list = copy.deepcopy(img_list)
        img_num = len(img_list)
        # Calculate the aspect ratio of all text bars
        width_list = []
        for img in img_list:
            width_list.append(img.shape[1] / float(img.shape[0]))
        # Sorting can speed up the cls process
        indices = np.argsort(np.array(width_list))

        cls_res = [['', 0.0]] * img_num
        batch_num = self.cls_batch_num
        elapse = 0
        for beg_img_no in range(0, img_num, batch_num):

            end_img_no = min(img_num, beg_img_no + batch_num)
            norm_img_batch = []
            max_wh_ratio = 0
            starttime = time.time()
            for ino in range(beg_img_no, end_img_no):
                h, w = img_list[indices[ino]].shape[0:2]
                wh_ratio = w * 1.0 / h
                max_wh_ratio = max(max_wh_ratio, wh_ratio)
            for ino in range(beg_img_no, end_img_no):
                norm_img = self.resize_norm_img(img_list[indices[ino]])
                norm_img = norm_img[np.newaxis, :]
                norm_img_batch.append(norm_img)
            norm_img_batch = np.concatenate(norm_img_batch)
            norm_img_batch = norm_img_batch.copy()

            if self.use_onnx:
                input_dict = {}
                input_dict[self.input_tensor.name] = norm_img_batch
                outputs = self.predictor.run(self.output_tensors, input_dict)
                prob_out = outputs[0]
            else:
                self.input_tensor.copy_from_cpu(norm_img_batch)
                self.predictor.run()
                prob_out = self.output_tensors[0].copy_to_cpu()
                self.predictor.try_shrink_memory()
            cls_result = self.postprocess_op(prob_out)
            elapse += time.time() - starttime
            for rno in range(len(cls_result)):
                label, score = cls_result[rno]
                cls_res[indices[beg_img_no + rno]] = [label, score]
                if '180' in label and score > self.cls_thresh:
                    img_list[indices[beg_img_no + rno]] = cv2.rotate(
                        img_list[indices[beg_img_no + rno]], 1)
        return img_list, cls_res, elapse


def main(args):
    image_file_list = get_image_file_list(args.image_dir)
    text_classifier = TextClassifier(args)
    valid_image_file_list = []
    img_list = []
    for image_file in image_file_list:
        img, flag = check_and_read_gif(image_file)
        if not flag:
            img = cv2.imread(image_file)
        if img is None:
            logger.info("error in loading image:{}".format(image_file))
            continue
        valid_image_file_list.append(image_file)
        img_list.append(img)
    try:
        img_list, cls_res, predict_time = text_classifier(img_list)
    except Exception as E:
        logger.info(traceback.format_exc())
        logger.info(E)
        exit()
    for ino in range(len(img_list)):
        logger.info("Predicts of {}:{}".format(valid_image_file_list[ino],
                                               cls_res[ino]))


if __name__ == "__main__":
    main(utility.parse_args())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2023/08/26

# from pyxllib.prog.pupil import check_install_package

# check_install_package('IPython', 'ipython')

import os
import sys
from multiprocessing import Process, Pipe
from io import StringIO
import threading

# from IPython.core.interactiveshell import InteractiveShell

# from timeout_decorator import timeout

from pyxllib.prog.pupil import format_exception
from pyxllib.file.specialist import XlPath

if sys.platform == 'win32':
    def timeout(seconds):
        def decorator(func):
            def wrapper(*args, **kwargs):
                result = None
                exception = None

                # 定义一个内部函数来运行目标函数
                def worker():
                    nonlocal result, exception
                    try:
                        result = func(*args, **kwargs)
                    except Exception as e:
                        exception = e

                thread = threading.Thread(target=worker)
                thread.start()
                thread.join(timeout=seconds)

                if thread.is_alive():
                    thread.join()  # 确保线程结束
                    raise TimeoutError(f"Function '{func.__name__}' exceeded {seconds} seconds of execution time.")
                if exception:
                    raise exception
                return result

            return wrapper

        return decorator


class InProcessExecutor:
    def __init__(self, base_dir=None):
        """ 注意为了应急，这里暂时是很不工程化的写法，一旦初始化这个类，工作目录、stdout、stderr 就被改变了 """
        self.base_dir = XlPath(base_dir) if base_dir else None
        self.user_ns = {}

        # 保存原始的工作目录、stdout 和 stderr
        self._original_cwd = os.getcwd()
        self._original_stdout = sys.stdout
        self._original_stderr = sys.stderr

        if self.base_dir:
            self.base_dir.mkdir(parents=True, exist_ok=True)
            os.chdir(self.base_dir)

    def _execute_code(self, code):
        response = {}
        try:
            # 捕获输出
            old_stdout, old_stderr = sys.stdout, sys.stderr
            redirected_output = StringIO()
            sys.stdout, sys.stderr = redirected_output, redirected_output

            # 使用 exec 执行代码
            exec(code, self.user_ns)

            # 恢复原始的 stdout 和 stderr
            sys.stdout, sys.stderr = old_stdout, old_stderr

            response['output'] = redirected_output.getvalue().strip()

        except Exception as e:
            response['error'] = format_exception(e)

        return response

    # @timeout(10)  # 先把超时写死了，以后有空再研究怎么加到函数的timeout参数里
    def execute(self, code: str, silent=False):
        """ 在主进程中执行代码 """
        response = self._execute_code(code)

        if not silent and response.get('output'):
            print(response['output'])

        if response.get('error'):
            print(f"执行错误: {response['error']}")

        return response

    def get_var(self, var_name: str):
        """ 获取指定的变量值 """
        return self.user_ns.get(var_name, None)

    def set_var(self, var_name: str, value):
        """ 设置变量的值 """
        self.user_ns[var_name] = value

    def close(self):
        """ 恢复原始的工作目录、stdout 和 stderr """
        os.chdir(self._original_cwd)
        sys.stdout = self._original_stdout
        sys.stderr = self._original_stderr


class SubprocessExecutor:
    """ 子进程版本 """

    def __init__(self, base_dir=None):
        """ 初始化执行器并启动子进程 """
        self.parent_conn, self.child_conn = Pipe()
        self.base_dir = XlPath(base_dir)
        self.process = Process(target=self._child_process, args=(self.child_conn, base_dir))
        self.process.start()

    def _child_process(self, conn, base_dir):
        """ 子进程主循环 """
        if base_dir:
            os.chdir(base_dir)

        user_ns = {}  # 初始化一个空的用户命名空间来替代 InteractiveShell 的 user_ns

        while True:
            message = conn.recv()  # 从父进程接收消息
            if message.get('action') == 'exit':
                break
            elif message.get('action') == 'execute':
                code = message.get('code')
                response = self._execute_code(code, user_ns)  # 传递user_ns为参数
                conn.send(response)
            elif message.get('action') == 'get_var':
                var_name = message.get('name')
                response = {'output': str(user_ns.get(var_name, None))}
                conn.send(response)
            elif message.get('action') == 'set_var':
                var_name = message.get('name')
                value = message.get('value')
                user_ns[var_name] = value

    def _execute_code(self, code, user_ns):
        response = {}
        try:
            # Capture output
            old_stdout, old_stderr = sys.stdout, sys.stderr
            redirected_output = StringIO()
            sys.stdout, sys.stderr = redirected_output, redirected_output

            # 使用 exec 来执行代码
            exec(code, user_ns)  # 使用传入的用户命名空间来执行代码

            # Restore original stdout and stderr
            sys.stdout, sys.stderr = old_stdout, old_stderr

            response['output'] = redirected_output.getvalue().strip()

        except Exception as e:
            response['error'] = format_exception(e)

        return response

    # @timeout(10)
    def execute(self, code: str, silent=False):
        """ 在子进程中执行代码 """
        self.parent_conn.send({'action': 'execute', 'code': code})
        response = self.parent_conn.recv()

        # 如果不处于静默模式，则处理响应
        if not silent:
            if response.get('output'):
                print(response['output'], end='')
            if response.get('error'):
                print(f"Error: {response['error']}")

        return response

    def get_var(self, var_name: str):
        """ 获取指定的环境变量值 """
        self.parent_conn.send({'action': 'get_var', 'name': var_name})
        response = self.parent_conn.recv()
        return response['output']

    def set_var(self, var_name: str, value):
        """ 设置环境变量的值 """
        self.parent_conn.send({'action': 'set_var', 'name': var_name, 'value': value})

    def close(self):
        """ 关闭子进程 """
        self.parent_conn.send({'action': 'exit'})
        self.process.join()


if os.getppid() == os.getpid():  # 主进程的时候，另外建立子进程运行更好，也便于调试
    AutoProcessExecutor = SubprocessExecutor
else:  # 并发跑的时候，只能在每个进程里，不能另外再套进程
    AutoProcessExecutor = InProcessExecutor


class InteractiveExecutor(SubprocessExecutor):
    """ 这个类是比较定制化的，还不算通用工具类

    但对我来说，暂时会经常用到，所以就写到这里来了
    """

    def execute(self, code: str, silent=False):
        """ 在子进程中执行代码

        相比父类，运行前会先记录当前文件清单，运行后会记录新的文件清单，然后对比两个文件清单，找出新的文件
        1. 为了简化问题，暂时只考虑工作目录下的直接文件，不考虑子目录，也不考虑对同名文件可能进行的修改
        2. 假设操作都只会在工作目录里，不会对工作目录外的文件使用绝对路径等手动操作
        """
        # 1 记录运行前的文件清单
        files1 = {f.name for f in self.base_dir.glob_files()}

        # 2 执行代码
        self.parent_conn.send({'action': 'execute', 'code': code})
        response = self.parent_conn.recv()

        # 3 记录运行后的文件清单
        files2 = {f.name for f in self.base_dir.glob_files()}
        out_files = list(files2 - files1)  # 相减后视为新增的文件
        if out_files:
            response['out_files'] = out_files

        # 如果不处于静默模式，则处理响应
        if not silent:
            if response['output']:
                print(response['output'])
            if 'error' in response:
                print(f"Error: {response['error']}")

        return response


if __name__ == '__main__':
    # 读取指定工作目录里的文件；只在指定工作目录里创建文件
    ie = SubprocessExecutor(XlPath.desktop())
    ie.execute('with open("a.txt", "r", encoding="utf8") as f: print(f.read())')
    # Hello
    r = ie.execute('with open("b.txt", "w", encoding="utf8") as f: f.write ("hello123")')
    print(r)  # {'output': '', 'files': {'b.txt'}}

    print(ie.get_var('test'))  # 索引不存在时返回None



# Copyright (c) 2021 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import yaml
from argparse import ArgumentParser, RawDescriptionHelpFormatter
import os.path
import logging
logging.basicConfig(level=logging.INFO)

support_list = {
    'it': 'italian',
    'xi': 'spanish',
    'pu': 'portuguese',
    'ru': 'russian',
    'ar': 'arabic',
    'ta': 'tamil',
    'ug': 'uyghur',
    'fa': 'persian',
    'ur': 'urdu',
    'rs': 'serbian latin',
    'oc': 'occitan',
    'rsc': 'serbian cyrillic',
    'bg': 'bulgarian',
    'uk': 'ukranian',
    'be': 'belarusian',
    'te': 'telugu',
    'ka': 'kannada',
    'chinese_cht': 'chinese tradition',
    'hi': 'hindi',
    'mr': 'marathi',
    'ne': 'nepali',
}

latin_lang = [
    'af', 'az', 'bs', 'cs', 'cy', 'da', 'de', 'es', 'et', 'fr', 'ga', 'hr',
    'hu', 'id', 'is', 'it', 'ku', 'la', 'lt', 'lv', 'mi', 'ms', 'mt', 'nl',
    'no', 'oc', 'pi', 'pl', 'pt', 'ro', 'rs_latin', 'sk', 'sl', 'sq', 'sv',
    'sw', 'tl', 'tr', 'uz', 'vi', 'latin'
]
arabic_lang = ['ar', 'fa', 'ug', 'ur']
cyrillic_lang = [
    'ru', 'rs_cyrillic', 'be', 'bg', 'uk', 'mn', 'abq', 'ady', 'kbd', 'ava',
    'dar', 'inh', 'che', 'lbe', 'lez', 'tab', 'cyrillic'
]
devanagari_lang = [
    'hi', 'mr', 'ne', 'bh', 'mai', 'ang', 'bho', 'mah', 'sck', 'new', 'gom',
    'sa', 'bgc', 'devanagari'
]
multi_lang = latin_lang + arabic_lang + cyrillic_lang + devanagari_lang

assert (os.path.isfile("./rec_multi_language_lite_train.yml")
        ), "Loss basic configuration file rec_multi_language_lite_train.yml.\
You can download it from \
https://github.com/PaddlePaddle/PaddleOCR/tree/dygraph/configs/rec/multi_language/"

global_config = yaml.load(
    open("./rec_multi_language_lite_train.yml", 'rb'), Loader=yaml.Loader)
project_path = os.path.abspath(os.path.join(os.getcwd(), "../../../"))


class ArgsParser(ArgumentParser):
    def __init__(self):
        super(ArgsParser, self).__init__(
            formatter_class=RawDescriptionHelpFormatter)
        self.add_argument(
            "-o", "--opt", nargs='+', help="set configuration options")
        self.add_argument(
            "-l",
            "--language",
            nargs='+',
            help="set language type, support {}".format(support_list))
        self.add_argument(
            "--train",
            type=str,
            help="you can use this command to change the train dataset default path"
        )
        self.add_argument(
            "--val",
            type=str,
            help="you can use this command to change the eval dataset default path"
        )
        self.add_argument(
            "--dict",
            type=str,
            help="you can use this command to change the dictionary default path"
        )
        self.add_argument(
            "--data_dir",
            type=str,
            help="you can use this command to change the dataset default root path"
        )

    def parse_args(self, argv=None):
        args = super(ArgsParser, self).parse_args(argv)
        args.opt = self._parse_opt(args.opt)
        args.language = self._set_language(args.language)
        return args

    def _parse_opt(self, opts):
        config = {}
        if not opts:
            return config
        for s in opts:
            s = s.strip()
            k, v = s.split('=')
            config[k] = yaml.load(v, Loader=yaml.Loader)
        return config

    def _set_language(self, type):
        lang = type[0]
        assert (type), "please use -l or --language to choose language type"
        assert(
                lang in support_list.keys() or lang in multi_lang
               ),"the sub_keys(-l or --language) can only be one of support list: \n{},\nbut get: {}, " \
                 "please check your running command".format(multi_lang, type)
        if lang in latin_lang:
            lang = "latin"
        elif lang in arabic_lang:
            lang = "arabic"
        elif lang in cyrillic_lang:
            lang = "cyrillic"
        elif lang in devanagari_lang:
            lang = "devanagari"
        global_config['Global'][
            'character_dict_path'] = 'ppocr/utils/dict/{}_dict.txt'.format(lang)
        global_config['Global'][
            'save_model_dir'] = './output/rec_{}_lite'.format(lang)
        global_config['Train']['dataset'][
            'label_file_list'] = ["train_data/{}_train.txt".format(lang)]
        global_config['Eval']['dataset'][
            'label_file_list'] = ["train_data/{}_val.txt".format(lang)]
        global_config['Global']['character_type'] = lang
        assert (
            os.path.isfile(
                os.path.join(project_path, global_config['Global'][
                    'character_dict_path']))
        ), "Loss default dictionary file {}_dict.txt.You can download it from \
https://github.com/PaddlePaddle/PaddleOCR/tree/dygraph/ppocr/utils/dict/".format(
            lang)
        return lang


def merge_config(config):
    """
    Merge config into global config.
    Args:
        config (dict): Config to be merged.
    Returns: global config
    """
    for key, value in config.items():
        if "." not in key:
            if isinstance(value, dict) and key in global_config:
                global_config[key].update(value)
            else:
                global_config[key] = value
        else:
            sub_keys = key.split('.')
            assert (
                sub_keys[0] in global_config
            ), "the sub_keys can only be one of global_config: {}, but get: {}, please check your running command".format(
                global_config.keys(), sub_keys[0])
            cur = global_config[sub_keys[0]]
            for idx, sub_key in enumerate(sub_keys[1:]):
                if idx == len(sub_keys) - 2:
                    cur[sub_key] = value
                else:
                    cur = cur[sub_key]


def loss_file(path):
    assert (
        os.path.exists(path)
    ), "There is no such file:{},Please do not forget to put in the specified file".format(
        path)


if __name__ == '__main__':
    FLAGS = ArgsParser().parse_args()
    merge_config(FLAGS.opt)
    save_file_path = 'rec_{}_lite_train.yml'.format(FLAGS.language)
    if os.path.isfile(save_file_path):
        os.remove(save_file_path)

    if FLAGS.train:
        global_config['Train']['dataset']['label_file_list'] = [FLAGS.train]
        train_label_path = os.path.join(project_path, FLAGS.train)
        loss_file(train_label_path)
    if FLAGS.val:
        global_config['Eval']['dataset']['label_file_list'] = [FLAGS.val]
        eval_label_path = os.path.join(project_path, FLAGS.val)
        loss_file(eval_label_path)
    if FLAGS.dict:
        global_config['Global']['character_dict_path'] = FLAGS.dict
        dict_path = os.path.join(project_path, FLAGS.dict)
        loss_file(dict_path)
    if FLAGS.data_dir:
        global_config['Eval']['dataset']['data_dir'] = FLAGS.data_dir
        global_config['Train']['dataset']['data_dir'] = FLAGS.data_dir
        data_dir = os.path.join(project_path, FLAGS.data_dir)
        loss_file(data_dir)

    with open(save_file_path, 'w') as f:
        yaml.dump(
            dict(global_config), f, default_flow_style=False, sort_keys=False)
    logging.info("Project path is          :{}".format(project_path))
    logging.info("Train list path set to   :{}".format(global_config['Train'][
        'dataset']['label_file_list'][0]))
    logging.info("Eval list path set to    :{}".format(global_config['Eval'][
        'dataset']['label_file_list'][0]))
    logging.info("Dataset root path set to :{}".format(global_config['Eval'][
        'dataset']['data_dir']))
    logging.info("Dict path set to         :{}".format(global_config['Global'][
        'character_dict_path']))
    logging.info("Config file set to       :configs/rec/multi_language/{}".
                 format(save_file_path))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2021/06/08 22:53

"""
TODO 写一些图片相似度相关功能
"""

from pyxllib.prog.pupil import check_install_package

check_install_package('imagehash', 'ImageHash')

import imagehash
import numpy as np

from pyxllib.cv.xlpillib import xlpil


def get_init_hash():
    """ 获得一个初始、空哈希值 """
    return imagehash.ImageHash(np.zeros([8, 8]).astype(bool))


def phash(image, *args, **kwargs):
    """ 修改了官方接口，这里输入的image支持泛用格式
    """
    im = xlpil.read(image)
    return imagehash.phash(im, *args, **kwargs)


def dhash(image, *args, **kwargs):
    """ 修改了官方接口，这里输入的image支持泛用格式

    官方比较推荐使用，性能速度又快的，就是dhash
    """
    im = xlpil.read(image)
    return imagehash.dhash(im, *args, **kwargs)



# copyright (c) 2021 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import cv2
import numpy as np
import pyclipper
from shapely.geometry import Polygon

__all__ = ['MakePseGt']


class MakePseGt(object):
    def __init__(self, kernel_num=7, size=640, min_shrink_ratio=0.4, **kwargs):
        self.kernel_num = kernel_num
        self.min_shrink_ratio = min_shrink_ratio
        self.size = size

    def __call__(self, data):

        image = data['image']
        text_polys = data['polys']
        ignore_tags = data['ignore_tags']

        h, w, _ = image.shape
        short_edge = min(h, w)
        if short_edge < self.size:
            # keep short_size >= self.size
            scale = self.size / short_edge
            image = cv2.resize(image, dsize=None, fx=scale, fy=scale)
            text_polys *= scale

        gt_kernels = []
        for i in range(1, self.kernel_num + 1):
            # s1->sn, from big to small
            rate = 1.0 - (1.0 - self.min_shrink_ratio) / (self.kernel_num - 1
                                                          ) * i
            text_kernel, ignore_tags = self.generate_kernel(
                image.shape[0:2], rate, text_polys, ignore_tags)
            gt_kernels.append(text_kernel)

        training_mask = np.ones(image.shape[0:2], dtype='uint8')
        for i in range(text_polys.shape[0]):
            if ignore_tags[i]:
                cv2.fillPoly(training_mask,
                             text_polys[i].astype(np.int32)[np.newaxis, :, :],
                             0)

        gt_kernels = np.array(gt_kernels)
        gt_kernels[gt_kernels > 0] = 1

        data['image'] = image
        data['polys'] = text_polys
        data['gt_kernels'] = gt_kernels[0:]
        data['gt_text'] = gt_kernels[0]
        data['mask'] = training_mask.astype('float32')
        return data

    def generate_kernel(self,
                        img_size,
                        shrink_ratio,
                        text_polys,
                        ignore_tags=None):
        """
        Refer to part of the code:
        https://github.com/open-mmlab/mmocr/blob/main/mmocr/datasets/pipelines/textdet_targets/base_textdet_targets.py
        """

        h, w = img_size
        text_kernel = np.zeros((h, w), dtype=np.float32)
        for i, poly in enumerate(text_polys):
            polygon = Polygon(poly)
            distance = polygon.area * (1 - shrink_ratio * shrink_ratio) / (
                polygon.length + 1e-6)
            subject = [tuple(l) for l in poly]
            pco = pyclipper.PyclipperOffset()
            pco.AddPath(subject, pyclipper.JT_ROUND, pyclipper.ET_CLOSEDPOLYGON)
            shrinked = np.array(pco.Execute(-distance))

            if len(shrinked) == 0 or shrinked.size == 0:
                if ignore_tags is not None:
                    ignore_tags[i] = True
                continue
            try:
                shrinked = np.array(shrinked[0]).reshape(-1, 2)
            except:
                if ignore_tags is not None:
                    ignore_tags[i] = True
                continue
            cv2.fillPoly(text_kernel, [shrinked.astype(np.int32)], i + 1)
        return text_kernel, ignore_tags
